struct Vec
    x :: Float64
    y :: Float64
    z :: Float64
end

function +(x :: Float64, y :: Vec)
    Vec(x + y.x, x + y.y, x + y.z)
end
function +(a :: Vec, b :: Float64)
    Vec(a.x + b, a.y + b, a.z + b)
end
function -(a :: Float64, y :: Vec)
    Vec(a + y.x, a + y.y, a + y.z)
end
function -(a :: Vec, b :: Float64)
    Vec(a.x - b, a.y - b, a.z - b)
end
function *(a :: Float64, b :: Vec)
    Vec(a * b.x, a * b.y, a * b.z)
end
function *(a :: Vec, b :: Float64)
    Vec(a.x*b, a.y*b, a.z*b)
end
function /(a :: Float64, b :: Vec)
    Vec(a/b.x, a/b.y, a/b.z)
end
function /(a :: Vec, b :: Float64)
    Vec(a.x/b, a.y/b, a.z/b)
end
function ^(a :: Float64, b :: Vec)
    Vec(a^b.x, a^b.y, a^b.z)
end
function ^(a :: Vec, b :: Float64)
    Vec(a.x^b, a.y^b, a.z^b)
end

function +(a :: Vec, b :: Vec)
    Vec(a.x+b.x, a.y + b.y, a.z+b.z)
end
function -(a :: Vec, b :: Vec)
    Vec(a.x-b.x, a.y-b.y, a.z-b.z)
end
function *(a :: Vec, b :: Vec)
    Vec(a.x*b.x, a.y*b.y, a.z*b.z)
end
function /(a :: Vec, b :: Vec)
    Vec(a.x/b.x, a.y/b.y, a.z/b.z)
end
function ^(a :: Vec, b :: Vec)
    Vec(a.x^b.x, a.y^b.y, a.z^b.z)
end

function -(a :: Vec)
    Vec(-a.x, -b.y, -b.z)
end

function abs(a :: Vec)
    Vec(abs(a.x), abs(a.y), abs(a.z))
end
function acos(a :: Vec)
    Vec(acos(a.x), acos(a.y), acos(a.z))
end
function acosh(a :: Vec)
    Vec(acosh(a.x), acosh(a.y), acosh(a.z))
end
function asin(a :: Vec)
    Vec(asin(a.x), asin(a.y), asin(a.z))
end
function atan(a :: Vec)
    Vec(atan(a.x), atan(a.y), atan(a.z))
end
function asinh(a :: Vec)
    Vec(asinh(a.x), asinh(a.y), asinh(a.z))
end
function atanh(a :: Vec)
    Vec(atanh(a.x), atanh(a.y), atanh(a.z))
end

function ceil(a :: Vec)
    Vec(ceil(a.x), ceil(a.y), ceil(a.z))
end
function clamp(a :: Float64, min :: Float64, max :: Float64)
    if a < min return min end
    if a > max return max end
    return a
end
function clamp(a :: Vec, min :: Float64, max :: Float64)
    Vec(clamp(a.x, min, max), clamp(a.y, min, max), clamp(a.z, min, max))
end
function clamp(a :: Vec, min :: Vec, max :: Vec)
    Vec(clamp(a.x, min.x, max.x), clamp(a.y, min.y, max.y), clamp(a.z, min.z, max.z))
end
function cos(a :: Vec)
    Vec(cos(a.x), cos(a.y), cos(a.z))
end
function cosh(a :: Vec)
    Vec(cosh(a.x), cosh(a.y), cosh(a.z))
end
function cross(a :: Vec, b :: Vec)
    Vec(a.y*b.z - b.y*a.z, a.z*b.x - b.z*a.x, a.x*b.y - b.x*b.y)
end

function distance(a :: Vec, b :: Vec)
    length(a - b)
end
function dot(a :: Vec, b :: Vec)
    a.x*b.x + a.y*b.y + a.z*b.z
end

function faceforward(N :: Vec, I :: Vec, Nref :: Vec)
    if dot(Nref, I) return N else return -N end
end
function floor(a :: Vec)
    Vec(floor(a.x), floor(a.y), floor(a.z))
end
function inversesqrt(a :: Float64)
    1/sqrt(a)
end
function inversesqrt(a :: Vec)
    Vec(inversesqrt(a.x), inversesqrt(a.y), inversesqrt(a.z))
end

function length(a :: Vec)
    sqrt(dot(a, a))
end
function log(a :: Vec)
    Vec(log(a.x), log(a.y), log(a.z))
end
function log2(a :: Vec)
    Vec(log(a.x, 2), log(a.y, 2), log(a.z, 2))
end

function max(a :: Vec, b :: Vec)
    Vec(max(a.x, b.x), max(a.y, b.y), max(a.z, b.z))
end
function min(a :: Vec, b :: Vec)
    Vec(min(a.x, b.x), min(a.y, b.y), min(a.z, b.z))
end
function mix(a, b, q)
    a*(1-q) + b*q
end
function mod(a :: Vec, b :: Vec)
    Vec(mod(a.x, b.x), mod(a.y, b.y), mod(a.z, b.z))
end

function normalize(a :: Vec)
    a/length(a)
end

function reflect(I :: Vec, N :: Vec)
    I - 2*dot(N, I)*N
end
function refract(I :: Vec, N :: Vec, eta :: Float64)
    k = 1 - eta*eta*(1 - dot(N, I)*dot(N, I))
    if k < 0 return Vec(0, 0, 0) end
    return eta*I - (eta*dot(N,I) + sqrt(k))*N
end
function round(a :: Vec)
    Vec(round(a.x), round(a.y), round(a.z))
end

function sign(a :: Vec)
    Vec(sign(a.x), sign(a.y), sign(a.z))
end
function sin(a :: Vec)
    Vec(sin(a.x), sin(a.y), sin(a.z))
end
function sinh(a :: Vec)
    Vec(sinh(a.x), sinh(a.y), sinh(a.z))
end
function smoothstep(edge0, edge1, x :: Vec)
    t = clamp((x - edge0)/(edge - 1), 0, 1)
    return t*t*(3 - 2*t)
end
function sqrt(a :: Vec)
    Vec(sqrt(a.x), sqrt(a.y), sqrt(a.z))
end
function step(edge :: Float64, x :: Float64)
    if x < edge return 0
    else return 1 end
end
function step(edge :: Float64, x :: Vec)
    Vec(step(edge, x.x), step(edge, x.y), step(edge, x.z))
end
function step(edge :: Vec, x :: Vec)
    Vec(step(edge.x, x.x), step(edge.y, x.y), step(edge.z, x.z))
end

function tan(a :: Vec)
    Vec(tan(a.x), tan(a.y), tan(a.z))
end
function tanh(a :: Vec)
    Vec(tanh(a.x), tanh(a.y), tanh(a.z))
end
function trunc(a :: Vec)
    Vec(trunc(a.x), trunc(a.y), trunc(a.z))
end

struct Mat
    xx :: Float64
    xy :: Float64
    xz :: Float64
    x0 :: Float64
    yx :: Float64
    yy :: Float64
    yz :: Float64
    y0 :: Float64
    zx :: Float64
    zy :: Float64
    zz :: Float64
    z0 :: Float64
end
function *(a :: Vec, b :: Mat)
    Vec(a.x*b.xx + a.y*b.xy + a.z*b.xz + b.x0,
        a.x*b.yx + a.y*b.yy + a.z*b.yz + b.y0,
        a.x*b.zx + a.y*b.zy + a.z*b.zz + b.z0)
end
# f(g(x))
# g(x).x = x.x*g.xx + x.y*g.xy + g.x0
# g(x).y = x.x*g.yx + x.y*g.yy + g.y0
# f(g(x)).x = g(x).x*f.xx + g(x).y*f.xy + f.x0
#       = (x.x*g.xx + x.y*g.xy + g.x0)*f.xx +
#           (x.x*g.yx + x.y*g.yy + g.y0)*f.xy + f.x0
#       = x.x*g.xx*f.xx + x.y*g.xy*f.xx + g.x0*f.xx + x.x*g.yx*f.xy +
#           x.y*g.yy*f.xy + g.y0*f.xy + f.x0
#       = x.x*(g.xx*f.xx + g.yx*f.xy) + x.y*(g.xy*f.xx + g.yy*f.xy) +
#           (g.x0*f.xx + g.y0*f.xy + f.x0)
# f(g(x)).y = g(x).x*f.yx + g(x).y*f.yy + f.y0
#       = (x.x*g.xx + x.y*g.xy + g.x0)*f.yx +
#           (x.x*g.yx + x.y*g.yy + g.y0)*f.yy + f.y0
#       = x.x*g.xx*f.yx + x.y*g.xy*f.yx + g.x0*f.yx +
#           x.x*g.yx*f.yy + x.y*g.yy*f.yy + g.y0*f.yy + f.y0
#       = x.x(g.xx*f.yx + g.yx*f.yy) + x.y*(g.xy*f.yx + g.yy*f.yy) +
#           (g.x0*f.yx + g.y0*f.yy + f.y0)
function *(f :: Mat, g :: Mat)
    Mat(g.xx*f.xx + g.yx*f.xy + g.zx*f.xz,
        g.xy*f.xx + g.yy*f.xy + g.zy*f.xz,
        g.xz*f.xx + g.yz*f.xy + g.zz*f.xz,
        g.x0*f.xx + g.y0*f.xy + g.z0*f.xz + f.x0,

        g.xx*f.yx + g.yx*f.yy + g.zx*f.yz,
        g.xy*f.yx + g.yy*f.yy + g.zy*f.yz,
        g.xz*f.yx + g.yz*f.yy + f.zz*f.yz,
        g.x0*f.yx + g.y0*f.yy + g.z0*f.yz + f.y0,

        g.xx*f.zx + g.yx*f.zy + g.zx*f.zz,
        g.xy*f.zx + g.yy*f.zy + g.zy*f.zz,
        g.xz*f.zx + g.yz*f.zy + g.zz*f.zz,
        g.x0*f.zx + g.y0*f.zy + g.z0*f.zz + f.z0)
end

# Not in the video...
function matrixCompMult(f :: Mat, g :: Mat)
    Mat(f.xx*g.xx, f.xy*g.xy, f.xz*g.xz, f.x0*g.x0,
        f.yx*g.yx, f.yy*g.yy, f.yz*g.yz, f.y0*g.y0,
        f.zx*g.zx, f.zy*g.zy, f.zz*g.zz, f.z0*g.z0)
end

# Yeah, I am not able to explain why this works...
# Based loosely on old Mesa3D code whilst hardcoding a 4th row.
# Might be buggy...
function inverse(m :: Mat)
    inv = Mat(m.yy*m.zz*1 - m.yy*m.z0*0 - m.zy*m.yz*1 + m.zy*m.y0*0 +
                 0*m.yz*m.z0 - 0*m.y0*m.zz,
              -m.xy*m.zz*1 + m.xy*m.z0*0 + m.zy*m.xz*1 - m.zy*m.x0*0 -
                 0*m.xz*m.z0 + 0*m.x0*m.zz,
              m.xy*m.yz*1 - m.xy*m.y0*0 - m.yy*m.xz*1 + m.yy*m.x0*0 +
                 0*m.xz*m.y0 - 0*m.x0*m.yz,
              -m.xy*m.yz*m.z0 + m.xy*m.y0*m.zz + m.yy*m.xz*m.z0 - m.yy*m.x0*m.zz - 
                 m.zy*m.xz*m.y0 + m.zy*m.x0*m.yz,

              -m.yx*m.zz*1 + m.yx*m.z0*0 + m.zx*m.yz*1 - m.zx*m.y0*0 - 
                 0*m.yz*m.z0 + 0*m.y0*m.zz,
              m.xx*m.zz*1 - m.xx*m.z0*0 - m.zx*m.xz*1 + m.zx*m.x0*0 + 
                 0*m.xz*m.z0 - 0*m.x0*m.zz,
              -m.xx*m.yz*1 + m.xx*m.y0*0 + m.yx*m.xz*1 - m.yx*m.x0*0 - 
                 0*m.xz*m.y0 + 0*m.x0*m.yz,
              m.xx*m.yz*m.z0 - m.xx*m.y0*m.zz - m.yx*m.xz*m.z0 + m.yx*m.x0*m.zz + 
                 m.zx * m.xz * m.y0 - m.zx * m.x0 * m.yz,

              m.yx*m.zy*1 - m.yx*m.z0*0 - m.zx*m.yy*1 + m.zx*m.y0*0 + 
                 0*m.yy*m.z0 - 0*m.y0*m.zy,
              -m.xx*m.zy*1 + m.xx*m.z0*0 + m.zx*m.xy*1 - m.zx*m.x0*0 - 
                 0*m.xy*m.z0 + 0*m.x0*m.zy,
              m.xx*m.yy*1 - m.xx*m.y0*0 - m.yx*m.xy*1 + m.yx*m.x0*0 + 
                 0*m.xy*m.y0 - 0*m.x0*m.yy,
              -m.xx*m.yy*m.z0 + m.xx*m.y0*m.zy + m.yx*m.xy*m.z0 - m.yx*m.x0*m.zy - 
                 m.zx*m.xy*m.y0 + m.zx*m.x0*m.yy)
    det = m.xx * inv.xx + m.xy * inv.yx + m.xz * inv.zx + m.x0 * 0
    if det == 0 return nothing end
    return Mat(m.xx/det, m.xy/det, m.xz/det, m.x0/det,
                m.yx/det, m.yy/det, m.yz/det, m.y0/det,
                m.zx/det, m.zy/det, m.zz/det, m.z0/det)
end
